
from .common import Iterable, Mapping, Sequence \
    , Complex, Integral, Number, Rational, Real \
    , hasws, stripsplit, uniques \
    , reduce, dbg_print
import warnings


RequiredValue = type("RequiredValueType", (object,), {})()

CONFIG_VALUE_TYPES = (bool, int, float, str \
    , type(RequiredValue), type(None))



class Setting(object):
  
  @classmethod
  def Factory(cls, name, tag, cast, defval=None, base=None, repr_=None):
    if repr_ is None:
      repr_ = cast.__repr__
    bases = (cls,) if base is None else (cls, base)
    defval, required = (None, True) if defval is RequiredValue else \
        (defval, False)
    mixins = \
        { "Tag": property(lambda s: tag, None, None, "setting tag")
        , "cast": property(lambda s: cast, None, None, "setting type")
        , "defval": property(lambda s: defval, None, None, "default value")
        , "required": property(lambda s: required, None, None, "required value")
        , "__repr__": repr_
        }
    return type(name, bases, mixins)
  
  _name     = None
  _value    = None
  
  @property
  def Tag(self):
    raise NotImplementedError("Tag on abstract class go figure.")
  
  @property
  def cast(self):
    raise NotImplementedError("cast on abstract class go figure.")
  
  @property
  def defval(self):
    raise NotImplementedError("name on abstract class go figure.")
  
  @property
  def required(self):
    raise NotImplementedError("required on abstract class go figure.")
  
  @property
  def name(self):
    return self._name
  
  @property
  def value(self):
    return self._value
  
  @value.setter
  def value(self, value):
    if value is None:
      if self.required:
        raise ValueError("{:s} requires an {:s} value" \
            .format(self._name, self.cast.__name__))
      value = self.defval
    self._value = None if value is None else self.cast(value)
  
  def __init__(self, name, value):
    self._name    = name
    self.value    = value
  
  def __bool__(self):
    return bool(self.defval if self._value is None else self._value)
  
  def __float__(self):
    return float(self.defval if self._value is None else self._value)
  
  def __int__(self):
    return int(self.defval if self._value is None else self._value)
  
  def __str__(self):
    return str(self.defval if self._value is None else self._value)

Setting.__index__ = Setting.__int__


class Configuration(Mapping):
  
  _settings     = None
  
  def __init__(self):
    pass





#!/usr/bin/env python

"""Tests for `yaspop` package."""

import pytest

from yaspop import common

def text_constants():
  """Test submodule common's constants."""
  assert type(None) is common.NoneType
  assert common.RequiredValue \
      and type(common.RequiredValue) is common.RequiredValueType

def test_required():
  """Test submodule common's required functions."""
  assert common.isrequired(common.RequiredValue)
  assert not common.isrequired(None)
  assert common.withdefault(None) is None
  assert common.withdefault(1) == 1
  assert common.withdefault(None, None) is None
  assert common.withdefault(1, None) == 1
  assert common.withdefault(None, 1) == 1
  assert common.withdefault(1, 1) == 1
  assert common.withdefault(1, types=int) == 1
  
  #assert withdefault(None, 1, types=None, name="obj") == 1

def test_izzy_functions():
  """Test submodule common's izzy functions."""
  assert common.isbool(bool(1))
  assert common.isfloat(float(1))
  assert common.isint(int(1))
  assert common.isstr(str(1))
  assert common.isnumeric(False)
  assert common.isnumeric(True)
  assert common.isnumeric(0)
  assert common.isnumeric(1)
  assert common.isnumeric(2.0)
  assert not common.isnumeric("2.0")


def test_isnseq():
  """Test submodule common's isnseq."""
  assert common.isnseq([], 0)
  assert common.isnseq([None], 1)
  assert common.isnseq([1,2], 2)
  assert common.isnseq([], 0, c=list)
  assert common.isnseq([None], 1, c=list)
  assert common.isnseq([1,2], 2, c=list)


def test_isnmseq():
  """Test submodule common's isnmseq."""
  assert common.isnmoseq((), 0)
  assert common.isnmoseq((None,), 1)
  assert common.isnmoseq((1,2), 2)
  assert common.isnmoseq((), 0, 2)
  assert common.isnmoseq((None,), 1, 2)
  assert common.isnmoseq((1,2), 2, 3)


def test_is2tuple():
  """Test submodule common's is2tuple."""
  assert common.is2tuple((None,None))
  assert common.is2tuple((None,1))
  assert common.is2tuple((0,None))
  assert common.is2tuple((1,2))
  assert common.is2tuple((0.5,1.0))
  assert common.is2tuple(("a","b"))


def test_isntuple():
  """Test submodule common's isntuple."""
  assert common.isntuple((), 0)
  assert common.isntuple((None,), 1)
  assert common.isntuple((1,2), 2)
  assert common.isntuple((), 0)
  assert common.isntuple((None,), 1)
  assert common.isntuple((1,2), 2)


def test_hasws():
  """Test submodule common's hasws."""
  assert common.hasws("all the time")
  assert common.hasws(" ")
  assert common.hasws("\n")
  assert not common.hasws("")


def test_stripsplit():
  """Test submodule common's stripsplit."""
  assert tuple(common.stripsplit("all the time")) == ("all", "the", "time")
  assert tuple(common.stripsplit("")) == ()
  assert tuple(common.stripsplit("   ")) == ()
  assert tuple(common.stripsplit(" many   differnt   things ")) == ("many", "differnt", "things")
  assert tuple(common.stripsplit(",,,,", ",")) == ()
  assert tuple(common.stripsplit("one, two, three", ",")) == ("one", "two", "three")
  assert tuple(common.stripsplit("one, two, three", ",", 1)) == ("one", "two, three")
  assert tuple(common.stripsplit("==one==, ==two==, three  ,=four===", ",", 2, '= ')) == ('one', 'two', 'three  ,=four')


def test_uniques():
  """Test submodule common's uniques."""
  assert common.uniques((1,2,3), (2,3,4), (3,4,5) * 10) == (1,2,3,4,5)
  assert common.uniques(tuple(range(20)), range(10,30), target=[]) == list(range(30))
  assert common.uniques((1,2,3), (2,2,3,4,4,4), (3,4,5,3,4,5,3,4,5), ()) == (1, 2, 3, 4, 5)
  assert common.uniques((1,2,3), (2,2,3,4,4,4), (3,4,5,3,4,5,3,4,5), (), target=[]) == [1, 2, 3, 4, 5]


def test_bind():
  """Test submodule common's bind."""
  assert common.bind(max, (-4,), (4,))() == 4
  assert common.bind(min, (-4,), (4,))() == -4


def test_dbg_print():
  """Test submodule common's dbg_print."""
  assert common.dbg_print('hello') is None


def test_classproperty():
  """Test classproperty decorator."""
  
  class Foo(object):
    @common.classproperty
    def z(cls):
      return "YES"
  
  f = Foo()
  assert Foo.z == "YES"
  assert f.z == "YES"



#!/usr/bin/env python

"""Tests for `yaspop` package."""

import pytest

from yaspop.common import RequiredValue
from yaspop.settings import Setting


def run_tests():
  class Foo():
    _settings = {}
    def __init__(self, *args):
      if args:
        self.s = args[0]
    def tag(self, name):
      return self._settings[name]._tag
  
  s = Setting('test')
  Foo.s = s
  f = Foo()
  assert f.s is None
  try:
    f.s = 8
    assert False
  except AttributeError as ae:
    pass
  
  s = Setting('test', RequiredValue)
  Foo.s = s
  try:
    f = Foo(None)
    assert False
  except AttributeError as ae:
    pass
  
  s = Setting('test', RequiredValue, setter=True)
  Foo.s = s
  try:
    f = Foo(None)
    assert False
  except AttributeError as ae:
    pass
  
  s = Setting('test', RequiredValue)
  Foo.s = s
  f = Foo(2)
  assert f.s == 2
  
  s = Setting('test', RequiredValue, setter=True)
  Foo.s = s
  f = Foo(3)
  assert f.s == 3
  
  assert bool(f.s) == True and isinstance(bool(f.s), bool)
  assert int(f.s) == 3 and isinstance(int(f.s), int)
  assert float(f.s) == 3.0 and isinstance(float(f.s), float)
  assert str(f.s) == "3" and isinstance(str(f.s), str)
  assert repr(f.s) == "3" and isinstance(repr(f.s), str)

def test_SettingFactory_bool():
  """Test submodule configurator's Setting.Factory."""
  """
  BoolSetting = Setting.Factory("test", "tag", bool)
  i = BoolSetting("test_bool", None)
  assert i.value is None
  i.value = 1
  assert i.value == True
  assert type(i.value) is bool
  i.value = "2"
  assert i.value == True
  assert type(i.value) is bool
  i.value = 3.
  assert i.value == True 
  assert type(i.value) is bool
  """

def test_SettingFactory_required_bool():
  """Test submodule configurator's Setting.Factory."""
  """
  BoolSetting = Setting.Factory("test", "tag", bool, defval=RequiredValue)
  try:
    i = BoolSetting("test_bool", None)
    assert False
  except:
    pass
  i = BoolSetting("test_bool", False)
  i.value = 1
  assert i.value == True
  assert type(i.value) is bool
  i.value = "2"
  assert i.value == True
  assert type(i.value) is bool
  i.value = 3.
  assert i.value == True 
  assert type(i.value) is bool
  """

def test_SettingFactory_default_bool():
  """Test submodule configurator's Setting.Factory."""
  """
  BoolSetting = Setting.Factory("test", "tag", bool, defval=False)
  i = BoolSetting("test_bool", None)
  assert i.value == False
  assert type(i.value) is bool
  i.value = 1
  assert i.value == True
  assert type(i.value) is bool
  i.value = "2"
  assert i.value == True
  assert type(i.value) is bool
  i.value = 3.
  assert i.value == True 
  assert type(i.value) is bool
  """

def test_SettingFactory_int():
  """Test submodule configurator's Setting.Factory."""
  """
  IntSetting = Setting.Factory("test", "tag", int)
  i = IntSetting("test_int", None)
  assert i.value is None
  i.value = 1
  assert i.value == 1
  assert type(i.value) is int
  i.value = "2"
  assert i.value == 2
  assert type(i.value) is int
  i.value = 3.
  assert i.value == 3 
  assert type(i.value) is int
  """

def test_SettingFactory_float():
  """Test submodule configurator's vreateConfiguration."""
  """
  IntSetting = Setting.Factory("test", "tag", float)
  i = IntSetting("test_float", None)
  assert i.value is None
  i.value = 1
  assert i.value == 1
  assert type(i.value) is float
  i.value = "2"
  assert i.value == 2
  assert type(i.value) is float
  i.value = 3.
  assert i.value == 3 
  assert type(i.value) is float
  """

def test_SettingFactory_str():
  """Test submodule configurator's vreateConfiguration."""
  """
  IntSetting = Setting.Factory("test", "tag", str)
  i = IntSetting("test_str", None)
  assert i.value is None
  i.value = 1
  assert i.value == "1"
  assert type(i.value) is str
  i.value = "2"
  assert i.value == "2"
  assert type(i.value) is str
  i.value = 3.
  assert i.value == "3.0" 
  assert type(i.value) is str
"""




from .common import Mapping, Sequence, RequiredValue \
    , dbg_print
import warnings


CONFIG_VALUE_TYPES = (bool, int, float, str \
    , type(RequiredValue), type(None))



class Setting(property):
  
  # class properties
  _tag      = 'abstract'
  _cast     = None
  
  @property
  def tag(cls):
    return self._tag
  
  @classmethod
  def Factory(cls, clsname, cast
      , bases=None, mixins=None, tag=None):
    if not isinstance(clsname, str):
      raise ValueError("clsname must be a non-empty string")
    if not callable(cast):
      raise ValueError("cast must be callable but is {:s}")
    if bases is None:
      bases = ()
    elif not isinstance(bases, Sequence) \
        or any(not isinstance(b, type) for b in bases):
      raise ValueError("bases must be a Sequence of classes or types")
    if mixins is None:
      mixins = {}
    elif not isinstance(mixins, Mapping):
      raise ValueError("mixins must be a Mapping of attributes")
    if tag is None:
      tag = clsname
    pcast = property(lambda s: tag, doc="cast {:s} to {:s}" \
        .format(tag, cast.__name__))
    return type(clsname, (Setting, *bases,)
        , {"cast": pcast, "tag": ptag, **mixins})
  
  def __new__(cls, name, default=None
      , getter=True, setter=None, deleter=None, doc=None
      ):
    self = super().__new__(cls)
    self._name = name
    return self
  
  # instance properties
  _default  = None
  
  @property
  def default(self):
    return self._default
  
  def __init__(self, name, default=None
      , getter=True, setter=None, deleter=None, doc="Undocumented"):
    dbg_print("Setting.__init__", name, default
        , getter, setter, deleter, doc)
    cast = self._cast
    value = None if default in (RequiredValue, None) else \
        cast(default)
    setter_wrapper = None
    if getter is True:
      if callable(setter):
        raise AttributeError("if {:s} getter is True then the".format(name) \
            + " setter may only be True or False (or None defaulting to True")
      if default is RequiredValue:
        if setter is False:
          raise AttributeError("if {:s} getter is True and".format(name) \
            + " default is RequiredValue then the setter may only be True" \
            + " (or None defaulting to True")
        if deleter:
          raise AttributeError("if {:s} getter is True and".format(name) \
            + " default is RequiredValue then deleters are not allowed")
      getter = lambda self: value
      if setter is True or default is RequiredValue:
        def setter_wrapper(self, val):
          nonlocal value, default
          if val is None:
            if default is RequiredValue:
              raise AttributeError("{:s} is a required value".format(name))
            val = None if default is None else cast(default)
          value = val
    elif setter is True:
      raise AttributeError("if {:s} getter is not True then the setter" \
      + " may only be callable or None".format(name))
    elif not setter and default is RequiredValue:
      raise AttributeError("if {:s} getter is not True and".format(name) \
          + " default is RequiredValue then the setter may only be callable")
    elif callable(setter):
      def setter_wrapper(self, val):
        nonlocal default, setter
        if val is None:
          if default is RequiredValue:
            raise AttributeError("{:s} is a required value".format(name))
          setter(self, None if default is None else cast(default))
    dbg_print("Setting.__init__ (before super)", name, default
      , getter, setter, deleter, doc)
    self._default = default
    super().__init__(getter, setter_wrapper, deleter, doc)
    dbg_print("Setting.__init__ (after super)", name, default
      , getter, setter, deleter, doc)
  
  def __bool__(self):
    return bool(self.default if self is None else self)
  
  def __float__(self):
    return float(self.default if self is None else self)
  
  def __int__(self):
    return int(self.default if self is None else self)
  
  def __str__(self):
    return str(self.default if self is None else self)
  
Setting.__index__ = Setting.__int__


class SettingMap(Mapping):
  
  @classmethod
  def Factory(cls, name, settings, base=None, repr_=None):
    # settings is (Setting instance, *setting_args)
    def __init__(self, initializer={}):
      for key in self._keys:
        self._settings[key] = settings[key](initializer[key] if key in initializer else None)
    def __getitem__(self, key):
      if key in self._keys:
        self._settings[key].value
      raise KeyError("invalid key {:s}".format(key))
    def __setitem__(self, key, value):
      if key in self._keys:
        self._settings[key].value = value
      raise KeyError("invalid key {:s}".format(key))
    bases = (cls,) if base is None else (cls, base)
    mixins = {"_settings": settings}
    return type(name, bases, mixins)




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
try:
  from collections.abc import Iterable, Mapping, Sequence
except ImportError:
  from collections import  Iterable, Mapping, Sequence

from numbers import Complex, Integral, Number, Rational, Real
from functools import reduce, partial

toseq = lambda obj, t=tuple: obj if isnseq(obj) else t((pbj,))

NoneType = type(None)
RequiredValueType = type("RequiredValueType", (object,), {})
RequiredValue = RequiredValueType()

isrequired = lambda obj: obj is RequiredValue

def withdefault(obj, default=None, types=None, name="obj"):
  if obj is None:
    if isrequired(default):
      raise ValueError("{:s} is required".format(name))
    obj = default
  if types is not None and not isinstance(obj, types):
    raise ValueError("{:s} expects {:s} but found {:s}" \
        .format(name
        , wordjoin(types, repr=lambda i: type(i).__name__)
        , repr(obj)))
  return obj

isbool = lambda obj: type(obj) is bool
isfloat = lambda obj: type(obj) is float
isint = lambda obj: type(obj) is int
isstr = lambda obj: type(obj) is str
isnumeric = lambda obj: isinstance(obj, Number)
isiterable = lambda obj: isinstance(obj, Iterable)

is2tuple = lambda obj, i=None: isnseq(obj, 2, c=tuple, i=i)
isntuple = lambda obj, n, i=None: isnseq(obj, n, c=tuple, i=i)
isnseq = lambda obj, n=0, c=Sequence, i=None: isnmoseq(obj, n, n, c, i)
isnmoseq = lambda obj, n=0, m=None, c=Sequence, i=None: \
    (c is None or isinstance(obj, c)) \
    and n <= len(obj) and (m is None or len(obj) <= m) \
    and (i is None or all(isinstance(i, i) for i in obj))

hasws   = lambda txt: isstr(txt) and any(c.isspace() for c in txt)

stripsplit = lambda txt, sep=None, maxsplit=-1, chars=None: \
    (t.strip(chars) for t in txt.split(sep, maxsplit) if t.strip(chars))

wordjoin = lambda seq, sep=', ', lastsep=' or ', repr=repr: \
    sep.join(repr(i) for i in seq[:-1]) \
    + (sep if lastsep is None else lastsep) + repr(seq[-1])

def uniques(*sources, target=None):
  ttype = type(sources[0] if target is None else target)
  if target is None:
    target = ttype(())
  for source in sources:
    target = reduce((lambda a,c: a if c in a else a+ttype((c,))) \
        , source, target)
  return target

def bind(func, pre=(), post=(), kwpre={}, kwpost={}):
  return lambda *args, **kwargs: \
    func(*(*pre,*args,*post), **{**kwpre,**kwargs,**kwpost})


class classproperty(property):
  def __get__(self, obj, objtype=None):
    return super().__get__(objtype)
  def __set__(self, obj, value):
    super().__set__(type(obj), value)
  def __delete__(self, obj):
    super().__delete__(type(obj))


if "DEBUG" not in globals():
  DEBUG = 0

def dbg_print(*args, **kargs):
  if DEBUG:
    print("DGB:", *args, **kargs)
